@@ -10,6 +10,7 @@ import (
 	"os"
 	"path/filepath"
 	"regexp"
+	"runtime"
 	"sort"
 	"strings"
 	"sync"
@@ -169,10 +170,11 @@ type IntegrationTestBuilder struct {
 	renamedFiles []string
 	renamedDirs  []string
 
-	buildCount int
-	GCCount    int
-	counters   *buildCounters
-	logBuff    lockingBuffer
+	buildCount   int
+	GCCount      int
+	counters     *buildCounters
+	logBuff      lockingBuffer
+	lastBuildLog string
 
 	builderInit sync.Once
 }
@@ -182,6 +184,13 @@ type lockingBuffer struct {
 	bytes.Buffer
 }
 
+func (b *lockingBuffer) ReadFrom(r io.Reader) (n int64, err error) {
+	b.Lock()
+	n, err = b.Buffer.ReadFrom(r)
+	b.Unlock()
+	return
+}
+
 func (b *lockingBuffer) Write(p []byte) (n int, err error) {
 	b.Lock()
 	n, err = b.Buffer.Write(p)
@@ -192,21 +201,21 @@ func (b *lockingBuffer) Write(p []byte) (n int, err error) {
 func (s *IntegrationTestBuilder) AssertLogContains(els ...string) {
 	s.Helper()
 	for _, el := range els {
-		s.Assert(s.logBuff.String(), qt.Contains, el)
+		s.Assert(s.lastBuildLog, qt.Contains, el)
 	}
 }
 
 func (s *IntegrationTestBuilder) AssertLogNotContains(els ...string) {
 	s.Helper()
 	for _, el := range els {
-		s.Assert(s.logBuff.String(), qt.Not(qt.Contains), el)
+		s.Assert(s.lastBuildLog, qt.Not(qt.Contains), el)
 	}
 }
 
 func (s *IntegrationTestBuilder) AssertLogMatches(expression string) {
 	s.Helper()
 	re := regexp.MustCompile(expression)
-	s.Assert(re.MatchString(s.logBuff.String()), qt.IsTrue, qt.Commentf(s.logBuff.String()))
+	s.Assert(re.MatchString(s.lastBuildLog), qt.IsTrue, qt.Commentf(s.lastBuildLog))
 }
 
 func (s *IntegrationTestBuilder) AssertBuildCountData(count int) {
@@ -281,7 +290,7 @@ func (s *IntegrationTestBuilder) AssertPublishDir(matches ...string) {
 func (s *IntegrationTestBuilder) AssertFs(fs afero.Fs, matches ...string) {
 	s.Helper()
 	var buff bytes.Buffer
-	helpers.PrintFs(fs, "", &buff)
+	s.Assert(s.printAndCheckFs(fs, "", &buff), qt.IsNil)
 	printFsLines := strings.Split(buff.String(), "\n")
 	sort.Strings(printFsLines)
 	content := strings.TrimSpace((strings.Join(printFsLines, "\n")))
@@ -304,6 +313,34 @@ func (s *IntegrationTestBuilder) AssertFs(fs afero.Fs, matches ...string) {
 	}
 }
 
+func (s *IntegrationTestBuilder) printAndCheckFs(fs afero.Fs, path string, w io.Writer) error {
+	if fs == nil {
+		return nil
+	}
+
+	return afero.Walk(fs, path, func(path string, info os.FileInfo, err error) error {
+		if err != nil {
+			return fmt.Errorf("error: path %q: %s", path, err)
+		}
+		path = filepath.ToSlash(path)
+		if path == "" {
+			path = "."
+		}
+		if !info.IsDir() {
+			f, err := fs.Open(path)
+			if err != nil {
+				return fmt.Errorf("error: path %q: %s", path, err)
+			}
+			defer f.Close()
+			// This will panic if the file is a directory.
+			var buf [1]byte
+			io.ReadFull(f, buf[:])
+		}
+		fmt.Fprintln(w, path, info.IsDir())
+		return nil
+	})
+}
+
 func (s *IntegrationTestBuilder) AssertFileExists(filename string, b bool) {
 	checker := qt.IsNil
 	if !b {
@@ -341,7 +378,7 @@ func (s *IntegrationTestBuilder) Build() *IntegrationTestBuilder {
 	s.Helper()
 	_, err := s.BuildE()
 	if s.Cfg.Verbose || err != nil {
-		fmt.Println(s.logBuff.String())
+		fmt.Println(s.lastBuildLog)
 		if s.H != nil && err == nil {
 			for _, s := range s.H.Sites {
 				m := s.pageMap
@@ -352,7 +389,7 @@ func (s *IntegrationTestBuilder) Build() *IntegrationTestBuilder {
 			}
 		}
 	} else if s.Cfg.LogLevel <= logg.LevelDebug {
-		fmt.Println(s.logBuff.String())
+		fmt.Println(s.lastBuildLog)
 	}
 	s.Assert(err, qt.IsNil)
 	if s.Cfg.RunGC {
@@ -364,7 +401,7 @@ func (s *IntegrationTestBuilder) Build() *IntegrationTestBuilder {
 }
 
 func (s *IntegrationTestBuilder) LogString() string {
-	return s.logBuff.String()
+	return s.lastBuildLog
 }
 
 func (s *IntegrationTestBuilder) BuildE() (*IntegrationTestBuilder, error) {
@@ -381,6 +418,7 @@ func (s *IntegrationTestBuilder) Init() *IntegrationTestBuilder {
 	if err := s.initBuilder(); err != nil {
 		s.Fatalf("Failed to init builder: %s", err)
 	}
+	s.lastBuildLog = s.logBuff.String()
 	return s
 }
 
@@ -626,10 +664,11 @@ func (s *IntegrationTestBuilder) build(cfg BuildCfg) error {
 	s.Helper()
 	defer func() {
 		s.reset()
+		s.lastBuildLog = s.logBuff.String()
+		s.logBuff.Reset()
 	}()
 
 	changeEvents := s.changeEvents()
-	s.logBuff.Reset()
 	s.counters = &buildCounters{}
 	cfg.testCounters = s.counters
 
@@ -643,16 +682,21 @@ func (s *IntegrationTestBuilder) build(cfg BuildCfg) error {
 	if err != nil {
 		return err
 	}
-	logErrorCount := s.H.NumLogErrors()
-	if logErrorCount > 0 {
-		return fmt.Errorf("logged %d error(s): %s", logErrorCount, s.logBuff.String())
-	}
 
 	return nil
 }
 
+// We simulate the fsnotify events.
+// See the test output in https://github.com/bep/fsnotifyeventlister for what events gets produced
+// by the different OSes.
 func (s *IntegrationTestBuilder) changeEvents() []fsnotify.Event {
-	var events []fsnotify.Event
+	var (
+		events    []fsnotify.Event
+		isLinux   = runtime.GOOS == "linux"
+		isMacOs   = runtime.GOOS == "darwin"
+		isWindows = runtime.GOOS == "windows"
+	)
+
 	for _, v := range s.removedFiles {
 		events = append(events, fsnotify.Event{
 			Name: v,
@@ -679,12 +723,32 @@ func (s *IntegrationTestBuilder) changeEvents() []fsnotify.Event {
 			Name: v,
 			Op:   fsnotify.Write,
 		})
+		if isLinux || isWindows {
+			// Duplicate write events, for some reason.
+			events = append(events, fsnotify.Event{
+				Name: v,
+				Op:   fsnotify.Write,
+			})
+		}
+		if isMacOs {
+			events = append(events, fsnotify.Event{
+				Name: v,
+				Op:   fsnotify.Chmod,
+			})
+		}
 	}
 	for _, v := range s.createdFiles {
 		events = append(events, fsnotify.Event{
 			Name: v,
 			Op:   fsnotify.Create,
 		})
+		if isLinux || isWindows {
+			events = append(events, fsnotify.Event{
+				Name: v,
+				Op:   fsnotify.Write,
+			})
+		}
+
 	}
 
 	// Shuffle events.
